@@ -10,28 +10,46 @@
 #include "customphrase.h"
 #include "symboldictionary.h"
 #include "workerthread.h"
+#include <cstddef>
+#include <cstdint>
 #include <fcitx-config/configuration.h>
+#include <fcitx-config/enum.h>
 #include <fcitx-config/iniparser.h>
 #include <fcitx-config/option.h>
+#include <fcitx-config/rawconfig.h>
 #include <fcitx-utils/event.h>
+#include <fcitx-utils/eventloopinterface.h>
+#include <fcitx-utils/handlertable.h>
 #include <fcitx-utils/i18n.h>
+#include <fcitx-utils/inputbuffer.h>
+#include <fcitx-utils/key.h>
+#include <fcitx-utils/keysym.h>
 #include <fcitx-utils/misc.h>
-#include <fcitx-utils/standardpath.h>
+#include <fcitx-utils/standardpaths.h>
+#include <fcitx-utils/trackableobject.h>
+#include <fcitx-utils/unixfd.h>
 #include <fcitx/action.h>
-#include <fcitx/addonfactory.h>
+#include <fcitx/addoninstance.h>
 #include <fcitx/addonmanager.h>
 #include <fcitx/event.h>
 #include <fcitx/inputcontext.h>
 #include <fcitx/inputcontextproperty.h>
 #include <fcitx/inputmethodengine.h>
 #include <fcitx/instance.h>
+#include <fcitx/text.h>
+#include <filesystem>
+#include <future>
 #include <libime/pinyin/pinyincontext.h>
 #include <libime/pinyin/pinyinime.h>
 #include <libime/pinyin/pinyinprediction.h>
+#include <list>
 #include <memory>
+#include <optional>
+#include <regex>
+#include <string>
 #include <string_view>
-#include <unordered_map>
-#include <unordered_set>
+#include <utility>
+#include <vector>
 
 namespace fcitx {
 
@@ -110,11 +128,13 @@ FCITX_CONFIGURATION(
     Option<bool> c{this, "C_CH", _("c <-> ch"), false};
     Option<bool> f{this, "F_H", _("f <-> h"), false};
     Option<bool> l{this, "L_N", _("l <-> n"), false};
+    Option<bool> r{this, "L_R", _("l <-> r"), false};
     Option<bool> s{this, "S_SH", _("s <-> sh"), false};
     Option<bool> z{this, "Z_ZH", _("z <-> zh"), false};
     OptionWithAnnotation<CorrectionLayout, CorrectionLayoutI18NAnnotation>
         correction{this, "Correction", _("Correction Layout"),
-                   CorrectionLayout::None};)
+                   isAndroid() ? CorrectionLayout::Qwerty
+                               : CorrectionLayout::None};)
 
 FCITX_CONFIGURATION(
     PinyinEngineConfig,
@@ -131,9 +151,15 @@ FCITX_CONFIGURATION(
     Option<bool> symbolsEnabled{this, "SymbolsEnabled", _("Enable Symbols"),
                                 true};
     Option<bool> chaiziEnabled{this, "ChaiziEnabled", _("Enable Chaizi"), true};
-    Option<bool> extBEnabled{this, "ExtBEnabled",
-                             _("Enable Characters in Unicode CJK Extension B"),
-                             !isAndroid()};
+    Option<bool> extBEnabled{
+        this, "ExtBEnabled",
+        _("Enable more Characters after Unicode CJK Extension B"),
+        !isAndroid()};
+    Option<bool> strokeCandidateEnabled{
+        this, "StrokeCandidateEnabled",
+        _("Show stroke candidates when typing with h(一), s(丨), p(丿), n(㇏), "
+          "z(𠃍)"),
+        true};
     OptionalHiddenSubConfigOption cloudpinyin{
         this, "CloudPinyin", _("Cloud Pinyin"),
         "fcitx://config/addon/cloudpinyin"};
@@ -155,7 +181,7 @@ FCITX_CONFIGURATION(
     Option<bool> preeditCursorPositionAtBeginning{
         this, "PreeditCursorPositionAtBeginning",
         _("Fix embedded preedit cursor at the beginning of the preedit"),
-        !isAndroid() && !isApple()};
+        !isAndroid() && !isApple() && !isEmscripten()};
     Option<bool> showActualPinyinInPreedit{
         this, "PinyinInPreedit", _("Show complete pinyin in preedit"), false};
     Option<bool> predictionEnabled{this, "Prediction", _("Enable Prediction"),
@@ -198,6 +224,20 @@ FCITX_CONFIGURATION(
         _("Next Candidate"),
         {Key("Tab")},
         KeyListConstrain({KeyConstrainFlag::AllowModifierLess})};
+    KeyListOption currentCandidate{
+        this,
+        "CurrentCandidate",
+        _("Select Current Candidate"),
+        {Key(FcitxKey_space), Key(FcitxKey_KP_Space)},
+        KeyListConstrain({KeyConstrainFlag::AllowModifierLess})};
+    KeyListOption commitRawInput{
+        this,
+        "CommitRawInput",
+        _("Commit Raw Input"),
+        {Key("Return"), Key("KP_Enter"), Key("Control+Return"),
+         Key("Control+KP_Enter"), Key("Shift+Return"), Key("Shift+KP_Enter"),
+         Key("Control+Shift+Return"), Key("Control+Shift+KP_Enter")},
+        KeyListConstrain({KeyConstrainFlag::AllowModifierLess})};
     KeyListOption secondCandidate{
         this,
         "SecondCandidate",
@@ -257,23 +297,52 @@ FCITX_CONFIGURATION(
     ExternalOption quickphrase{this, "QuickPhrase", _("Quick Phrase"),
                                "fcitx://config/addon/quickphrase/editor"};
     OptionWithAnnotation<std::vector<std::string>, ToolTipAnnotation>
-        quickphraseTrigger{this,
-                           "QuickPhrase trigger",
-                           _("Strings to trigger quick phrase"),
-                           {"www.", "ftp.", "http:", "mail.", "bbs.", "forum.",
-                            "https:", "ftp:", "telnet:", "mailto:"},
-                           {},
-                           {},
-                           {_("Enter a string from the list will make it enter "
-                              "quickphrase mode.")}};
+        quickphraseTriggerRegex{
+            this,
+            "QuickPhraseTriggerRegex",
+            _("Regular expression to trigger quick phrase"),
+            {".(/|@)$", "^(www|bbs|forum|mail|bbs)\\.",
+             "^(http|https|ftp|telnet|mailto):"},
+            {},
+            {},
+            {_("Enter quickphrase mode when current input matches any regular "
+               "expression from the list.")}};
     Option<FuzzyConfig> fuzzyConfig{this, "Fuzzy", _("Fuzzy Pinyin Settings")};
     HiddenOption<bool> firstRun{this, "FirstRun", "FirstRun", true};)
 
-class PinyinState;
 struct EventSourceTime;
 class CandidateList;
+class PinyinEngine;
 
-class PinyinEngine final : public InputMethodEngineV3 {
+enum class PinyinMode { Normal, StrokeFilter, ForgetCandidate, Punctuation };
+
+class PinyinState : public InputContextProperty {
+public:
+    PinyinState(PinyinEngine *engine);
+
+    libime::PinyinContext context_;
+    bool lastIsPunc_ = false;
+
+    PinyinMode mode_ = PinyinMode::Normal;
+
+    // Stroke filter
+    std::shared_ptr<CandidateList> strokeCandidateList_;
+    InputBuffer strokeBuffer_;
+
+    // Forget candidate
+    std::shared_ptr<CandidateList> forgetCandidateList_;
+
+    std::unique_ptr<EventSourceTime> cancelLastEvent_;
+
+    std::optional<std::vector<std::string>> predictWords_;
+
+    int keyReleased_ = -1;
+    int keyReleasedIndex_ = -2;
+    uint64_t lastKeyPressedTime_ = 0;
+};
+
+class PinyinEngine final : public InputMethodEngineV3,
+                           public TrackableObject<PinyinEngine> {
 public:
     PinyinEngine(Instance *instance);
     ~PinyinEngine() override;
@@ -317,6 +386,11 @@ public:
 
     void resetStroke(InputContext *inputContext) const;
     void resetForgetCandidate(InputContext *inputContext) const;
+    void forgetCandidate(InputContext *inputContext, size_t index);
+    void pinCustomPhrase(InputContext *inputContext,
+                         const std::string &customPhrase);
+    void deleteCustomPhrase(InputContext *inputContext,
+                            const std::string &customPhrase);
 
     FCITX_ADDON_DEPENDENCY_LOADER(cloudpinyin, instance_->addonManager());
 
@@ -329,11 +403,14 @@ private:
 
     bool handleCloudpinyinTrigger(KeyEvent &event);
     bool handle2nd3rdSelection(KeyEvent &event);
-    bool handleCandidateList(KeyEvent &event);
-    bool handleNextPage(KeyEvent &event);
-    bool handleStrokeFilter(KeyEvent &event);
+    bool handleCandidateList(KeyEvent &event,
+                             const std::shared_future<uint32_t> &keyChr);
+    bool handleNextPage(KeyEvent &event) const;
+    bool handleStrokeFilter(KeyEvent &event,
+                            const std::shared_future<uint32_t> &keyChr);
     bool handleForgetCandidate(KeyEvent &event);
-    bool handlePunc(KeyEvent &event);
+    bool handlePunc(KeyEvent &event,
+                    const std::shared_future<uint32_t> &keyChr);
     bool handlePuncCandidate(KeyEvent &event);
     bool handleCompose(KeyEvent &event);
     void resetPredict(InputContext *inputContext);
@@ -362,16 +439,16 @@ private:
     void loadBuiltInDict();
     void loadExtraDict();
     void loadCustomPhrase();
-    void loadSymbols(const StandardPathFile &file);
-    void loadDict(StandardPathFile file,
+    void loadSymbols(const UnixFD &file);
+    void loadDict(const std::string &fullPath,
                   std::list<std::unique_ptr<TaskToken>> &taskTokens);
+    void saveCustomPhrase();
 
     Instance *instance_;
     PinyinEngineConfig config_;
     PinyinEngineConfig pyConfig_;
     std::unique_ptr<libime::PinyinIME> ime_;
-    std::unordered_map<std::string, std::unordered_set<uint32_t>>
-        quickphraseTriggerDict_;
+    std::vector<std::regex> quickphraseTriggerRegex_;
     KeyList selectionKeys_;
     KeyList numpadSelectionKeys_;
     FactoryFor<PinyinState> factory_;
@@ -398,13 +475,6 @@ private:
     static constexpr size_t NumBuiltInDict = 2;
 };
 
-class PinyinEngineFactory : public AddonFactory {
-public:
-    AddonInstance *create(AddonManager *manager) override {
-        registerDomain("fcitx5-chinese-addons", FCITX_INSTALL_LOCALEDIR);
-        return new PinyinEngine(manager->instance());
-    }
-};
 } // namespace fcitx
 
 #endif // _PINYIN_PINYIN_H_
